SystemOrganization addCategory: #UnificationTests!

AlgebraicDataType subclass: #VariadicTerm
	instanceVariableNames: 'subterms'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'UnificationTests'!

!VariadicTerm commentStamp: 'fbs 5/29/2011 16:47' prior: 0!
A general term that may contain any (variable) number of subterms. It's mainly just for testing.!

!VariadicTerm class methodsFor: 'instance variables' stamp: 'fbs 5/27/2011 20:26'!
with: aCollection
	^ self new with: aCollection.! !

!VariadicTerm methodsFor: 'destructuring' stamp: 'fbs 5/27/2011 20:27'!
unapply
	^ #Term asOrderedCollection, subterms.! !

!VariadicTerm methodsFor: 'private' stamp: 'fbs 5/27/2011 20:26'!
with: aCollection
	subterms := aCollection! !

TestCase subclass: #MatchingTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'UnificationTests'!

!MatchingTest methodsFor: 'helpers' stamp: 'fbs 7/22/2011 07:55'!
assert: left equals: right description: msg
	self
		assert: left = right
		description: ('{1} , expected {2}, got {3}' format: {msg. left printString. right printString.}).! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 7/28/2011 14:18'!
testEmptyMguPassedIntoValueBlockMatchWins
	| result substitution |
	result := 1 matchOneOf: {[1] -> [:mgu | substitution := mgu. 'foo']} otherwise: ['bar'].
	self assert: result = 'foo'.
	self assert: substitution isEmpty.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:34'!
testFirstMatchWins
	| result |
	result := 1
		matchOneOf:
			{[1] -> ['first match'].
			[1] -> ['second match'].}
		otherwise: ['otherwise block'].
	self assert: result = 'first match' description: 'Unexpected match invoked'.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:35'!
testMatchPassesUnifierIntoBlock
	| result mgu |
	result := {1}
		matchOneOf:
			{[{#x asVariable}] -> [:vars | mgu := vars. 'match']}
		otherwise: ['otherwise block'].
	self assert: result = 'match' description: 'No match found'.
	self assert: mgu notNil description: 'MGU not passed into block'.
	self assert: (mgu at: #x asVariable) = 1 description: 'Unexpected MGU'.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:36'!
testMatchingTriesAllBlocks
	| result |
	result := 1
		matchOneOf:
			{[0] -> ['first match'].
			[1] -> ['second match'].}
		otherwise: ['otherwise block'].
	self assert: result = 'second match' description: 'Unexpected result'.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:32'!
testMatchingUnifiesBidirectionally
	| result mgu |
	result := {#x asVariable. 1}
		matchOneOf:
			{[{2. #y asVariable}] -> [:vars | mgu := vars. 'match']}
		otherwise: ['otherwise block'].
	self assert: result = 'match' description: 'No match found'.
	self assert: mgu notNil description: 'MGU not passed into block'.
	self assert: (mgu at: #x asVariable) equals: 2 description: 'Unexpected MGU (x)'.
	self assert: (mgu at: #y asVariable) equals: 1 description: 'Unexpected MGU (y)'.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:33'!
testNoMatchEvaluatesOtherwiseBlock
	| result |
	result := 1
		matchOneOf:
			{['not a 1'] -> ['matched something']}
		otherwise: ['otherwise block'].
	self assert: result = 'otherwise block' description: 'Otherwise block not invoked'.! !

!MatchingTest methodsFor: 'testing - matching' stamp: 'fbs 1/5/2012 09:33'!
testNoMatchEvaluatesRaisesError
	self should: [1
		matchOneOf:
			{['not a 1'] -> ['matched something']}]
		raise: Error.! !

TestCase subclass: #MetaVariableTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'UnificationTests'!

!MetaVariableTest methodsFor: 'testing - comparing' stamp: 'fbs 1/4/2012 20:34'!
testMetaVariableEqualToSameNamedMetaVariable
	self assert: #x asVariable equals: #x asVariable.! !

!MetaVariableTest methodsFor: 'testing - comparing' stamp: 'fbs 1/4/2012 20:35'!
testMetaVariableHashEqualToSameNamedMetaVariableHash
	self assert: #x asVariable hash equals: #x asVariable hash.! !

!MetaVariableTest methodsFor: 'testing - comparing' stamp: 'fbs 1/4/2012 20:36'!
testMetaVariableHashNotEqualToDifferentlyNamedMetaVariableHash
	self deny: #x asVariable hash = #X asVariable hash.! !

!MetaVariableTest methodsFor: 'testing - comparing' stamp: 'fbs 1/4/2012 20:35'!
testMetaVariableNotEqualToDifferentlyNamedMetaVariable
	self deny: #x asVariable = #X asVariable.! !

!MetaVariableTest methodsFor: 'testing - comparing' stamp: 'fbs 1/4/2012 20:35'!
testMetaVariableNotEqualToNil
	self deny: #x asVariable = nil! !

!MetaVariableTest methodsFor: 'testing' stamp: 'fbs 5/24/2011 11:03'!
testUnapplyShouldDestructureCorrectly
	self assert: (MetaVariable named: #a) unapply = {#Variable. #a.}! !

!MetaVariableTest methodsFor: 'testing' stamp: 'fbs 5/24/2011 22:08'!
testVariableHasNoChildren
	self deny: (MetaVariable named: #foo) hasChildren.

	self assert: (MetaVariable named: #foo) children isEmpty.! !

!MetaVariableTest methodsFor: 'testing' stamp: 'fbs 5/25/2011 22:32'!
testVariableIsVariable
	self assert: (MetaVariable named: #foo) isMetaVariable.! !

TestCase subclass: #OrUnifierTest
	instanceVariableNames: 'matcher'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'UnificationTests'!

!OrUnifierTest methodsFor: 'running' stamp: 'fbs 10/7/2011 07:45'!
setUp
	matcher := OrUnifier
		left: (TreeNode left: #x asVariable)
		right: (TreeNode right: #x asVariable).! !

!OrUnifierTest methodsFor: 'test' stamp: 'fbs 10/7/2011 07:45'!
testFirstOptionUnifies
	| mgu |
	mgu := matcher =? (TreeNode left: (Leaf value: 1)).
	self assert: (mgu at: (#x asVariable)) = (Leaf value: 1).! !

!OrUnifierTest methodsFor: 'test' stamp: 'fbs 8/9/2011 07:30'!
testNoUnificationPossibleStillSignalsError
	self should: [matcher =? Tree empty] raise: UnificationFailure! !

!OrUnifierTest methodsFor: 'test' stamp: 'fbs 10/7/2011 07:45'!
testSecondOptionUnifies
	| mgu |
	mgu := matcher =? (TreeNode right: (Leaf value: 1)).
	self assert: (mgu at: (#x asVariable)) = (Leaf value: 1).! !

!OrUnifierTest methodsFor: 'test' stamp: 'fbs 12/28/2011 09:30'!
testUnifiesAsRightHandOperandAgainstFirstOption
	| mgu |
	mgu := (TreeNode left: (Leaf value: 1)) =? matcher.
	self assert: (mgu at: (#x asVariable)) = (Leaf value: 1).! !

!OrUnifierTest methodsFor: 'test' stamp: 'fbs 12/28/2011 09:30'!
testUnifiesAsRightHandOperandAgainstSecondOption
	| mgu |
	mgu := (TreeNode right: (Leaf value: 1)) =? matcher.
	self assert: (mgu at: (#x asVariable)) = (Leaf value: 1).! !

TestCase subclass: #SolutionFinderTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'UnificationTests'!

!SolutionFinderTest commentStamp: 'fbs 1/4/2012 10:39' prior: 0!
Most of SolutionFinder's tests are found in UnificationTest.!

!SolutionFinderTest methodsFor: 'fixtures' stamp: 'fbs 1/5/2012 09:28'!
makePartition
	| partition |
	partition := VariableTrackingUnionFind usingArrayType: PersistentCollection partitioning: Dictionary new.
	partition := (partition find: 1) unificationClosureWith: (partition find: #x asVariable) in: partition.
	^ partition.! !

!SolutionFinderTest methodsFor: 'testing - instance creation' stamp: 'fbs 1/5/2012 09:29'!
testCanCreateFinder
	| partition |
	partition := self makePartition.
	SolutionFinder findSolutionFor: partition starting: 1.! !

TestCase subclass: #UnificationMatcherTest
	instanceVariableNames: 'matcher substitution'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'UnificationTests'!

!UnificationMatcherTest methodsFor: 'helpers' stamp: 'fbs 7/28/2011 22:17'!
assert: left equals: right description: msg
	self
		assert: left = right
		description: ('{1} , expected {2}, got {3}' format: {msg. left printString. right printString.}).! !

!UnificationMatcherTest methodsFor: 'running' stamp: 'fbs 10/7/2011 07:44'!
setUp
	matcher := UnificationMatcher matcherUsing:
		{[Leaf value: 1] -> ['first leaf'].
		[Leaf value: 1] -> ['second leaf'].
		[Tree empty] -> ['tree'].
		[TreeNode left: #x asVariable] -> [:mgu | substitution := mgu. 'mgu-using case']}
		otherwise: ['otherwise block'].! !

!UnificationMatcherTest methodsFor: 'testing - matching' stamp: 'fbs 8/8/2011 19:09'!
testFirstMatchWins
	| result |
	result := matcher match: (Leaf value: 1).
	self assert: result = 'first leaf' description: 'Unexpected match invoked'.! !

!UnificationMatcherTest methodsFor: 'testing - matching' stamp: 'fbs 10/7/2011 07:44'!
testMatchPassesUnifierIntoBlock
	| result |
	result := matcher match: (TreeNode left: #y asVariable).
	
	self assert: result = 'mgu-using case' description: 'No match found'.
	self assert: substitution notNil description: 'MGU not passed into block'.
	self assert: (substitution at: #y asVariable) = (#x asVariable) description: 'Unexpected MGU'.! !

!UnificationMatcherTest methodsFor: 'testing - matching' stamp: 'fbs 8/8/2011 20:26'!
testMatchingTriesAllBlocks
	| result |
	result := matcher match: Tree empty.
	self assert: result = 'tree' description: 'Unexpected result'.! !

!UnificationMatcherTest methodsFor: 'testing - matching' stamp: 'fbs 8/8/2011 20:26'!
testNoMatchEvaluatesOtherwiseBlock
	| result |
	result := matcher match: (Leaf value: 2).
	self assert: result = 'otherwise block' description: 'Otherwise block not invoked'.! !

!UnificationMatcherTest methodsFor: 'testing - matching' stamp: 'fbs 7/28/2011 22:28'!
testNoMatchEvaluatesRaisesError
	| m |
	m := UnificationMatcher matcherUsing:
			{[Tree empty] -> ['matched something']}.
			
	self should: [m matchAgainst: (Leaf value: 1)] raise: Error.! !

TestCase subclass: #UnificationOfAlgebraicDataTypeTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'UnificationTests'!

!UnificationOfAlgebraicDataTypeTest methodsFor: 'testing' stamp: 'fbs 1/5/2012 09:29'!
testAlgebraicDataTypeOrReturnsOrMatcher
	| mgu |
	mgu := ((TreeNode left: #x asVariable) or: (TreeNode right: #x asVariable)) =? (TreeNode right: (Leaf value: 1)).
	self assert: (mgu at: (#x asVariable)) = (Leaf value: 1).! !

!UnificationOfAlgebraicDataTypeTest methodsFor: 'testing' stamp: 'fbs 1/5/2012 09:24'!
testDifferentArityTermsShouldNotUnify
	"f(x) ~= g(x,y)"
	|f g |
	f := VariadicTerm with: #(#x).
	g := VariadicTerm with: #(#x #y).
	self should: [f unifyWith: g] raise: UnificationFailure.! !

!UnificationOfAlgebraicDataTypeTest methodsFor: 'testing - collection unification' stamp: 'fbs 1/9/2012 09:15'!
testNormalObjectsCantUnifyWithAlgebraicDataTypes
	self should: [(Leaf value: 1) =? #(Leaf 1)] raise: UnificationFailure.! !

!UnificationOfAlgebraicDataTypeTest methodsFor: 'testing - collection unification' stamp: 'fbs 1/5/2012 09:25'!
testUnificationWithCollectionsWorksBidirectionally
	| mgu |
	mgu := {#x asVariable. 1.} =? {VariadicTerm with: {#a asVariable}. #y asVariable.}.
	self assert: (VariadicTerm with: {#a asVariable}) equals: (mgu at: #x asVariable).
	self assert: 1 equals: (mgu at: #y asVariable).! !

TestCase subclass: #UnificationOfCollectionsTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'UnificationTests'!

!UnificationOfCollectionsTest methodsFor: 'testing' stamp: 'fbs 1/5/2012 09:26'!
testNormalObjectsDontUnifyWithCollections
	self should: [1 =? #()] raise: UnificationFailure.
	self should: [#() =? 1] raise: UnificationFailure.! !

!UnificationOfCollectionsTest methodsFor: 'testing' stamp: 'fbs 1/5/2012 09:26'!
testUnificationWorksBidirectionally
	| mgu |
	mgu := {#x asVariable. 1.} =? {2. #y asVariable}.
	self assert: 2 equals: (mgu at: #x asVariable).
	self assert: 1 equals: (mgu at: #y asVariable).! !

!UnificationOfCollectionsTest methodsFor: 'testing' stamp: 'fbs 10/7/2011 07:57'!
testUnificationWorksLeftward
	| mgu |
	mgu := {#x asVariable. #y asVariable.} =? #(1 (2 3)).
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: #(2 3) equals: (mgu at: #y asVariable).! !

!UnificationOfCollectionsTest methodsFor: 'testing' stamp: 'fbs 10/7/2011 07:56'!
testUnificationWorksRightward
	| mgu |
	mgu := #(1 (2 3)) =? {#x asVariable. #y asVariable.}.
	self assert: 1 equals: (mgu at: #x asVariable).
	self assert: #(2 3) equals: (mgu at: #y asVariable).! !

TestCase subclass: #UnificationTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'UnificationTests'!

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/8/2012 22:49'!
testAnythingCanUnifyAgainstMetaVariables
	self shouldnt: ['a' =? #x asVariable] raise: UnificationFailure.
	self shouldnt: [$a =? #x asVariable] raise: UnificationFailure.
	self shouldnt: [1 =? #x asVariable] raise: UnificationFailure.
	self shouldnt: [#() =? #x asVariable] raise: UnificationFailure.
	self shouldnt: [Object new =? #x asVariable] raise: UnificationFailure.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 22:16'!
testCyclicStructuresCantUnify
	| y |
	y := MetaVariable named: #y.
	self should: [y unifyWith: {y}] raise: UnificationFailure.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 22:16'!
testImpossibleUnificationShouldNotUnify
	self should: [1 unifyWith: 2] raise: UnificationFailure.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/8/2012 22:56'!
testMetaVariablesCanUnifyAgainstAnything
	self shouldnt: [#x asVariable =? 'a'] raise: UnificationFailure.
	self shouldnt: [#x asVariable =? $a] raise: UnificationFailure.
	self shouldnt: [#x asVariable =? 1] raise: UnificationFailure.
	self shouldnt: [#x asVariable =? #()] raise: UnificationFailure.
	self shouldnt: [#x asVariable =? Object new] raise: UnificationFailure.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 5/29/2011 15:08'!
testSelfUnificationShouldUnify
	| x |
	x := MetaVariable named: #x.
	"Tautologies"
	self assert: (1 unifyWith: 1) isEmpty.
	self assert: (x unifyWith: x) isEmpty.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 5/26/2011 07:43'!
testTrivialBidirectionalUnificationShouldUnify
	| matches x y |
	x := MetaVariable named: #x.
	y := MetaVariable named: #y.
	matches := x unifyWith: y.
	self assert: matches = (Dictionary newFrom: {x -> y}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 22:18'!
testTrivialUnificationsShouldUnify
	| x |
	x := MetaVariable named: #x.
	self assert: (x unifyWith: 2) = (Dictionary newFrom: {x -> 2}).
	self assert: (2 unifyWith: x) = (Dictionary newFrom: {x -> 2}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 22:22'!
testTwoVariableBidirectionalUnificationShouldUnifyBoth
	| matches left right |
	left := MetaVariable named: #left.
	right := MetaVariable named: #right.
	matches := {left. 2} unifyWith: {1. right}.
	self assert: matches = (Dictionary newFrom: {
		left -> 1.
		right -> 2}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 22:21'!
testTwoVariableUnificationShouldUnifyBoth
	| matches left right |
	left := MetaVariable named: #left.
	right := MetaVariable named: #right.
	matches := {left. right} unifyWith: {1. 2}.
	self assert: matches = (Dictionary newFrom: {
		left -> 1.
		right -> 2}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/5/2012 09:27'!
testUnificationWorksOnDeepStructures
	| matches x y |
	x := MetaVariable named: #x.
	y := MetaVariable named: #y.
	matches := { {{x}} } unifyWith: {y}.
	self assert: matches = (Dictionary newFrom: { y -> {{x}} }).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 22:24'!
testVariableUnifyingAgainstTwoVariables
	| matches x y z |
	x := MetaVariable named: #x.
	y := MetaVariable named: #y.
	z := MetaVariable named: #z.
	matches := {x. x.} unifyWith: {y. z.}.
	self assert: matches = (Dictionary newFrom: {y -> x. z -> x}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 1/4/2012 21:36'!
testVariablesInsideCollectionsUnify
	| matches |
	matches := {1. 2. #x asVariable. 4} =? #(1 2 3 4).
	self assert: matches equals: (Dictionary newFromPairs: {#x asVariable. 3.}).
	
	matches := #(1 2 3 4) =? {1. 2. #x asVariable. 4}.
	self assert: matches equals: (Dictionary newFromPairs: {#x asVariable. 3.}).! !

TestCase subclass: #VariableTrackingUnionFindTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'UnificationTests'!

!VariableTrackingUnionFindTest methodsFor: 'fixtures' stamp: 'fbs 1/4/2012 20:26'!
makePartition
	^ VariableTrackingUnionFind
		usingArrayType: PersistentCollection
		partitioning: Dictionary new! !

!VariableTrackingUnionFindTest methodsFor: 'as yet unclassified' stamp: 'fbs 1/4/2012 16:25'!
testAddingNonVariableToVariableClassMakesNonVariableTheRepresentative
	| v |
	v := self makePartition.
	v := v union: #x asVariable with: 1.
	self assert: 1 equals: (v find: 1).
	self assert: 1 equals: (v find: #x asVariable).! !

!VariableTrackingUnionFindTest methodsFor: 'as yet unclassified' stamp: 'fbs 1/4/2012 16:25'!
testAddingVariableToClassMakesNonVariableTheRepresentative
	| v |
	v := self makePartition.
	v := v union: 1 with: #x asVariable.
	self assert: 1 equals: (v find: 1).
	self assert: 1 equals: (v find: #x asVariable).! !

!VariableTrackingUnionFindTest methodsFor: 'as yet unclassified' stamp: 'fbs 1/4/2012 16:26'!
testAddingVariableToVariableClassMakesOriginalVariableTheRepresentative
	| v |
	v := self makePartition.
	v := v union: #x asVariable with: #y asVariable.
	self assert: #y asVariable equals: (v find: #y asVariable).
	self assert: #y asVariable equals: (v find: #x asVariable).! !
